@@ -12,20 +12,29 @@ import copy
 path = os.path.abspath(os.path.dirname(__file__))
 sys.path.append(path)
 
-from opensnitch import ui_pb2
-from opensnitch import ui_pb2_grpc
+import opensnitch.proto as proto
+ui_pb2, ui_pb2_grpc = proto.import_()
 
 from opensnitch.dialogs.prompt import PromptDialog
 from opensnitch.dialogs.stats import StatsDialog
 
+from opensnitch.plugins import PluginsManager, PluginBase, PluginsList, PluginSignal
+from opensnitch.actions import Actions
 from opensnitch.notifications import DesktopNotifications
 from opensnitch.firewall import Rules as FwRules
 from opensnitch.nodes import Nodes
 from opensnitch.config import Config
 from opensnitch.version import version
 from opensnitch.database import Database
-from opensnitch.utils import Utils, CleanerTask, Themes
-from opensnitch.utils import Message, languages
+from opensnitch.utils import (
+    Utils,
+    CleanerTask,
+    Themes,
+    OneshotTimer,
+    languages,
+    Message
+)
+from opensnitch.utils.duration import duration
 from opensnitch.utils.xdg import Autostart
 
 class UIService(ui_pb2_grpc.UIServicer, QtWidgets.QGraphicsObject):
@@ -35,8 +44,9 @@ class UIService(ui_pb2_grpc.UIServicer, QtWidgets.QGraphicsObject):
     _add_alert_trigger = QtCore.pyqtSignal(str, str, ui_pb2.Alert)
     _version_warning_trigger = QtCore.pyqtSignal(str, str)
     _status_change_trigger = QtCore.pyqtSignal(bool)
-    _notification_callback = QtCore.pyqtSignal(ui_pb2.NotificationReply)
+    _notification_callback = QtCore.pyqtSignal(str, ui_pb2.NotificationReply)
     _show_message_trigger = QtCore.pyqtSignal(str, str, int, int)
+    _temp_rule_expired = QtCore.pyqtSignal(str, str)
 
     # .desktop filename located under /usr/share/applications/
     DESKTOP_FILENAME = "opensnitch_ui.desktop"
@@ -92,6 +102,12 @@ class UIService(ui_pb2_grpc.UIServicer, QtWidgets.QGraphicsObject):
         self._remote_stats = {}
         self._autostart = Autostart()
 
+        # create plugins and actions before dialogs
+        self._plugin_mgr = PluginsManager.instance()
+        self._actions = Actions().instance()
+        self._plugin_mgr.load_plugins()
+        self._actions.loadAll()
+
         self.translator = None
         self._init_translation()
         self._themes = Themes()
@@ -114,6 +130,8 @@ class UIService(ui_pb2_grpc.UIServicer, QtWidgets.QGraphicsObject):
                 'ports':{},
                 'users':{}
                 }
+        # list of active scheduled tasks to disable temporal rules.
+        self._sched_tasks = {}
 
         if not start_in_bg:
             self._show_gui_if_tray_not_available()
@@ -128,6 +146,29 @@ class UIService(ui_pb2_grpc.UIServicer, QtWidgets.QGraphicsObject):
         if self._cfg.getBool(self._cfg.DEFAULT_IGNORE_RULES):
             self._nodes.delete_rule_by_field(Config.DURATION_FIELD, Config.RULES_DURATION_FILTER)
 
+        self._load_plugins()
+
+    def _load_plugins(self):
+        for action_conf in self._actions.getAll():
+            # get one global conf
+            action_def = self._actions.get(action_conf)
+            # initialize global plugins
+            if action_def == None or action_def.get('type') == None or PluginBase.TYPE_GLOBAL not in action_def['type']:
+                continue
+            # get the actions defined in the conf
+            for name in action_def['actions']:
+                try:
+                    action = action_def['actions'][name]
+                    action.run()
+                except Exception as e:
+                    print("global._load_plugins() exception:", e, "is {0} enabled?".format(name))
+
+    def _stop_plugins(self):
+        for plug in PluginsList.actions:
+            p = plug()
+            p.signal_in.emit({"plugin": p.get_name(), "signal": PluginSignal.STOP})
+        self._plugin_mgr.unload_all()
+
     # https://gist.github.com/pklaus/289646
     def _setup_interfaces(self):
         namestr, outbytes = Utils.get_interfaces()
@@ -151,6 +192,7 @@ class UIService(ui_pb2_grpc.UIServicer, QtWidgets.QGraphicsObject):
         self._stats_dialog.settings_saved.connect(self._on_settings_saved)
         self._stats_dialog.close_trigger.connect(self._on_close)
         self._show_message_trigger.connect(self._show_systray_message)
+        self._temp_rule_expired.connect(self._on_temp_rule_expired)
 
     def _setup_icons(self):
         self.off_image = QtGui.QPixmap(os.path.join(self._path, "res/icon-off.png"))
@@ -261,6 +303,7 @@ class UIService(ui_pb2_grpc.UIServicer, QtWidgets.QGraphicsObject):
     def _on_close(self):
         self._exit = True
         self._tray.setIcon(self.off_icon)
+        self._prompt_dialog.close()
         self._app.processEvents()
         self._nodes.stop_notifications()
         self._nodes.update_all(Nodes.OFFLINE)
@@ -268,6 +311,11 @@ class UIService(ui_pb2_grpc.UIServicer, QtWidgets.QGraphicsObject):
         self._db.optimize()
         self._db.close()
         self._stop_db_cleaner()
+        self._stop_plugins()
+
+        for k in self._sched_tasks:
+            self._sched_tasks[k].stop()
+
         self._on_exit()
 
     def _show_stats_dialog(self):
@@ -302,32 +350,30 @@ class UIService(ui_pb2_grpc.UIServicer, QtWidgets.QGraphicsObject):
 
     @QtCore.pyqtSlot(str, str, ui_pb2.Alert)
     def _on_new_alert(self, proto, addr, alert):
+        # TODO: move to its own module
         try:
             is_local = self._is_local_request(proto, addr)
 
-            icon = QtWidgets.QSystemTrayIcon.Information
-            _title = QtCore.QCoreApplication.translate("messages", "Info")
-            atype = "INFO"
-            if alert.type == ui_pb2.Alert.ERROR:
-                atype = "ERROR"
-                _title = QtCore.QCoreApplication.translate("messages", "Error")
-                icon = QtWidgets.QSystemTrayIcon.Critical
-            if alert.type == ui_pb2.Alert.WARNING:
-                atype = "WARNING"
-                _title = QtCore.QCoreApplication.translate("messages", "Warning")
-                icon = QtWidgets.QSystemTrayIcon.Warning
-
-            body = ""
             what = "GENERIC"
-            if alert.what == ui_pb2.Alert.GENERIC:
-                body = alert.text
-            elif alert.what == ui_pb2.Alert.KERNEL_EVENT:
+            body = alert.text
+            if alert.what == ui_pb2.Alert.KERNEL_EVENT:
                 body = "%s\n%s" % (alert.text, alert.proc.path)
                 what = "KERNEL EVENT"
             if is_local is False:
                 body = "node: {0}:{1}\n\n{2}\n{3}".format(proto, addr, alert.text, alert.proc.path)
 
             if alert.action == ui_pb2.Alert.SHOW_ALERT:
+                icon = QtWidgets.QSystemTrayIcon.Information
+                _title = QtCore.QCoreApplication.translate("messages", "Info")
+                atype = "INFO"
+                if alert.type == ui_pb2.Alert.ERROR:
+                    atype = "ERROR"
+                    _title = QtCore.QCoreApplication.translate("messages", "Error")
+                    icon = QtWidgets.QSystemTrayIcon.Critical
+                if alert.type == ui_pb2.Alert.WARNING:
+                    atype = "WARNING"
+                    _title = QtCore.QCoreApplication.translate("messages", "Warning")
+                    icon = QtWidgets.QSystemTrayIcon.Warning
 
                 urgency = DesktopNotifications.URGENCY_NORMAL
                 if alert.priority == ui_pb2.Alert.LOW:
@@ -337,6 +383,12 @@ class UIService(ui_pb2_grpc.UIServicer, QtWidgets.QGraphicsObject):
 
                 self._show_message_trigger.emit(_title, body, icon, urgency)
 
+                self._db.insert("alerts",
+                                "(time, node, type, action, priority, what, body, status)",
+                                (
+                                    datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
+                                    proto+":"+addr, atype, "", "", what, body, 0
+                                ))
             else:
                 print("PostAlert() unknown alert action:", alert.action)
 
@@ -365,8 +417,8 @@ class UIService(ui_pb2_grpc.UIServicer, QtWidgets.QGraphicsObject):
         else:
             self._tray.setIcon(self.off_icon)
 
-    @QtCore.pyqtSlot(ui_pb2.NotificationReply)
-    def _on_notification_reply(self, reply):
+    @QtCore.pyqtSlot(str, ui_pb2.NotificationReply)
+    def _on_notification_reply(self, addr, reply):
         if reply.code == ui_pb2.ERROR:
             self._tray.showMessage("Error",
                                 reply.data,
@@ -423,6 +475,8 @@ class UIService(ui_pb2_grpc.UIServicer, QtWidgets.QGraphicsObject):
         if self.translator:
             self._app.removeTranslator(self.translator)
         saved_lang = self._cfg.getSettings(Config.DEFAULT_LANGUAGE)
+        if saved_lang == languages.DEFAULT_LANG:
+            return
         self.translator = languages.init(saved_lang)
         self._app.installTranslator(self.translator)
 
@@ -669,6 +723,15 @@ class UIService(ui_pb2_grpc.UIServicer, QtWidgets.QGraphicsObject):
 
         return newconf
 
+    @QtCore.pyqtSlot(str, str)
+    def _on_temp_rule_expired(self, node_addr, rule_name):
+        self._nodes.disable_rule(node_addr, rule_name)
+        try:
+            key = node_addr+rule_name
+            del self._sched_tasks[key]
+        except:
+            pass
+
     @QtCore.pyqtSlot(dict)
     def _on_node_actions(self, kwargs):
         if kwargs['action'] == self.NODE_ADD:
@@ -688,8 +751,9 @@ class UIService(ui_pb2_grpc.UIServicer, QtWidgets.QGraphicsObject):
         elif kwargs['action'] == self.ADD_RULE:
             rule = kwargs['rule']
             proto, addr = self._get_peer(kwargs['peer'])
+            node_addr = "{0}:{1}".format(proto, addr)
             self._nodes.add_rule((datetime.now().strftime("%Y-%m-%d %H:%M:%S")),
-                                 "{0}:{1}".format(proto, addr),
+                                 node_addr,
                                  rule.name, rule.description, str(rule.enabled),
                                  str(rule.precedence), str(rule.nolog), rule.action, rule.duration,
                                  rule.operator.type, str(rule.operator.sensitive), rule.operator.operand,
@@ -701,12 +765,30 @@ class UIService(ui_pb2_grpc.UIServicer, QtWidgets.QGraphicsObject):
                 # daemon.
                 rule.operator.data = ""
 
+            # disable temporal rules in the db
+            def _disable_temp_rule(args):
+                self._temp_rule_expired.emit(args[0], args[1].name)
+
+            timeout = duration.to_seconds(rule.duration)
+            if rule.duration == Config.DURATION_ONCE:
+                timeout = 1
+            if timeout > 0:
+                ost = OneshotTimer(timeout, _disable_temp_rule, (node_addr, rule,))
+                key = node_addr + rule.name
+                self._sched_tasks[key] = ost
+                ost.start()
+
         elif kwargs['action'] == self.DELETE_RULE:
             self._db.delete_rule(kwargs['name'], kwargs['addr'])
 
         elif kwargs['action'] == self.NODE_DELETE:
             self._delete_node(kwargs['peer'])
 
+    def OpenWindow(self):
+        self._stats_dialog.show()
+
+    def close(self):
+        self._on_close()
 
     def PostAlert(self, alert, context):
         proto, addr = self._get_peer(context.peer())
@@ -746,8 +828,9 @@ class UIService(ui_pb2_grpc.UIServicer, QtWidgets.QGraphicsObject):
 
         # TODO: allow connections originated from ourselves: os.getpid() == request.pid)
         self._asking = True
-        proto, addr = self._get_peer(context.peer())
-        rule, timeout_triggered = self._prompt_dialog.promptUser(request, self._is_local_request(proto, addr), context.peer())
+        peer = context.peer()
+        proto, addr = self._get_peer(peer)
+        rule, timeout_triggered = self._prompt_dialog.promptUser(request, self._is_local_request(proto, addr), peer)
         self._last_ping = datetime.now()
         self._asking = False
         if rule == None:
@@ -766,20 +849,19 @@ class UIService(ui_pb2_grpc.UIServicer, QtWidgets.QGraphicsObject):
                                             QtWidgets.QSystemTrayIcon.NoIcon,
                                             DesktopNotifications.URGENCY_NORMAL)
 
-
         if rule.duration in Config.RULES_DURATION_FILTER:
             self._node_actions_trigger.emit(
                 {
                     'action': self.DELETE_RULE,
                     'name': rule.name,
-                    'addr': context.peer()
+                    'addr': peer
                 }
             )
         else:
             self._node_actions_trigger.emit(
                 {
                     'action': self.ADD_RULE,
-                    'peer': context.peer(),
+                    'peer': peer,
                     'rule': rule
                 }
             )
@@ -868,6 +950,7 @@ class UIService(ui_pb2_grpc.UIServicer, QtWidgets.QGraphicsObject):
                 try:
                     if stop_event.is_set():
                         break
+
                     in_message = next(node_iter)
                     if in_message == None:
                         continue
@@ -899,6 +982,3 @@ class UIService(ui_pb2_grpc.UIServicer, QtWidgets.QGraphicsObject):
                 context.cancel()
 
         return node_iter
-
-    def OpenWindow(self):
-        self._stats_dialog.show()
